@@ -23,9 +23,7 @@ import Distribution.Simple.UUAGC.AbsSyn( AGFileOption(..)
 import Distribution.Simple.UUAGC.Parser
 import Options hiding (verbose)
 import Distribution.Verbosity
-import System.Process( CreateProcess(..), createProcess, CmdSpec(..)
-                     , StdStream(..), runProcess, waitForProcess
-                     , shell)
+import System.Process( readProcessWithExitCode )
 import System.Directory(getModificationTime
                        ,doesFileExist
                        ,removeFile)
@@ -90,22 +88,15 @@ uuagcUserHook' uuagcPath = uuagcLibUserHook (uuagcFromString uuagcPath)
 -- | Create uuagc function using shell (old method)
 uuagcFromString :: String -> [String] -> FilePath -> IO (ExitCode, [FilePath])
 uuagcFromString uuagcPath args file = do
-  let argline = uuagcPath ++ concatMap (' ':) (args ++ [file])
-  (_, Just ppOutput, Just ppError, ph) <- createProcess (shell argline)
-                                    { std_in  = Inherit
-                                    , std_out = CreatePipe
-                                    , std_err = CreatePipe
-                                    }
-  ec <- waitForProcess ph
+  (ec,out,err) <- readProcessWithExitCode uuagcPath (args ++ [file]) ""
   case ec of
     ExitSuccess ->
-      do putErrorInfo ppError
-         fls <- processContent ppOutput
-         return (ExitSuccess, fls)
+      do hPutStr stderr err
+         return (ExitSuccess, words out)
     (ExitFailure exc) ->
       do hPutStrLn stderr (uuagcPath ++ ": " ++ show exc)
-         putErrorInfo ppOutput
-         putErrorInfo ppError
+         hPutStr stderr out
+         hPutStr stderr err
          return (ExitFailure exc, [])
 
 -- | Main hook, argument should be uuagc function
@@ -121,9 +112,6 @@ originalPreBuild  = preBuild simpleUserHooks
 originalBuildHook = buildHook simpleUserHooks
 originalSDistHook = sDistHook simpleUserHooks
 
-processContent :: Handle -> IO [String]
-processContent = liftM words . hGetContents
-
 putErrorInfo :: Handle -> IO ()
 putErrorInfo h = hGetContents h >>= hPutStr stderr
 
@@ -137,29 +125,36 @@ updateAGFile :: ([String] -> FilePath -> IO (ExitCode, [FilePath]))
              -> IO ()
 updateAGFile _ _ (_,(_,Nothing)) = return ()
 updateAGFile uuagc newOptions (file,(opts,Just (gen,sp))) = do
-  (ec, files) <- uuagc (optionsToString $ opts { genFileDeps = True, searchPath = sp }) file
-  case ec of
-    ExitSuccess ->
-      do when ((not.null) files) $ do
+  hasGen <- doesFileExist gen
+  when hasGen $ do
+    (ec, files) <- uuagc (optionsToString $ opts { genFileDeps = True, searchPath = sp }) file
+    case ec of
+      ExitSuccess -> do
+        let newOpts :: Options 
+            newOpts = maybe noOptions fst $ Map.lookup file newOptions
+            optRebuild = optionsToString newOpts /= optionsToString opts
+        modRebuild <-
+          if null files
+          then return False
+          else do
             flsmt <- mapM getModificationTime files
             let maxModified = maximum flsmt
             fmt <- getModificationTime gen
-            let newOpts :: Options 
-                newOpts = maybe noOptions fst $ Map.lookup file newOptions
-            -- When some dependency is newer or options have changed, we should regenerate
-            when (maxModified > fmt || optionsToString newOpts /= optionsToString opts) $ removeFile gen
-    ex@(ExitFailure _) -> throwIO ex
+            return $ maxModified > fmt
+        -- When some dependency is newer or options have changed, we should regenerate
+        when (optRebuild || modRebuild) $ removeFile gen
+      ex@(ExitFailure _) -> throwIO ex
 
 getAGFileOptions :: [(String, String)] -> IO AGFileOptions
 getAGFileOptions extra = do
+  cabalOpts <- mapM (parseOptionAG . snd) $ filter ((== agModule) . fst) extra
   usesOptionsFile <- doesFileExist defUUAGCOptions
   if usesOptionsFile
        then do r <- parserAG' defUUAGCOptions
                case r of
-                 Left e -> print e >> exitFailure
-                 Right a -> return a
-       else mapM (parseOptionAG . snd)
-            $ filter ((== agModule) . fst) extra
+                 Left e -> dieNoVerbosity (show e)
+                 Right a -> return $ cabalOpts ++ a
+       else return cabalOpts
 
 getAGClasses :: [(String, String)] -> IO [AGOptionsClass]
 getAGClasses = mapM (parseClassAG . snd) . filter ((== agClass) . fst)
@@ -256,25 +251,27 @@ commonHook uuagc classesPath pd lbi fl = do
 getAGFileList :: AGFileOptions -> [FilePath]
 getAGFileList = map (normalise . filename)
 
-uuagc :: BuildInfo -> LocalBuildInfo -> PreProcessor
+uuagc :: BuildInfo -> LocalBuildInfo -> ComponentLocalBuildInfo -> PreProcessor
 uuagc = uuagc' (uuagcFromString uuagcn)
 
 uuagc' :: ([String] -> FilePath -> IO (ExitCode, [FilePath]))
         -> BuildInfo
         -> LocalBuildInfo
+        -> ComponentLocalBuildInfo
         -> PreProcessor
-uuagc' uuagc build lbi  =
+uuagc' uuagc build lbi _ =
    PreProcessor {
      platformIndependent = True,
      runPreProcessor = mkSimplePreProcessor $ \ inFile outFile verbosity ->
-                       do putStrLn $ "[UUAGC] processing: " ++ inFile ++ " generating: " ++ outFile
+                       do notice verbosity $ "[UUAGC] processing: " ++ inFile ++ " generating: " ++ outFile
                           let classesPath = buildDir lbi </> agClassesFile
                           info verbosity $ "uuagc-preprocessor: Assuming AG classesPath: " ++ classesPath
                           fileOpts <- readFileOptions classesPath
-                          let opts = case Map.lookup inFile fileOpts of
-                                       Nothing        -> noOptions
-                                       Just (opt,gen) -> opt
-                              search  = dropFileName inFile
+                          opts <- case Map.lookup inFile fileOpts of
+                                       Nothing        -> do warn verbosity $ "No options found for " ++ inFile
+                                                            return noOptions
+                                       Just (opt,gen) -> return opt
+                          let search  = dropFileName inFile
                               options = opts { searchPath = search : hsSourceDirs build ++ searchPath opts
                                              , outputFiles = outFile : (outputFiles opts) }
                           (eCode,_) <- uuagc (optionsToString options) inFile
@@ -283,8 +280,8 @@ uuagc' uuagc build lbi  =
                             ex@(ExitFailure _) -> throwIO ex
                 }
 
-nouuagc :: BuildInfo -> LocalBuildInfo -> PreProcessor
-nouuagc build lbi =
+nouuagc :: BuildInfo -> LocalBuildInfo -> ComponentLocalBuildInfo -> PreProcessor
+nouuagc build lbi _ =
   PreProcessor {
     platformIndependent = True,
     runPreProcessor = mkSimplePreProcessor $ \inFile outFile verbosity -> do
